import dataclasses
from typing import Literal

from django.db import connection


@dataclasses.dataclass
class RelationSpaceStat:
    schema: str
    name: str
    type: Literal["table", "index"]
    size_bytes: int
    row_estimate: int


def get_database_size() -> int:
    """Returns the size of the current database in bytes"""
    with connection.cursor() as cursor:
        cursor.execute(
            """
            SELECT pg_database_size(current_database()) as database_size
        """
        )
        return cursor.fetchone()[0]


def get_relation_sizes() -> list[RelationSpaceStat]:
    with connection.cursor() as cursor:
        cursor.execute(
            """
SELECT
    n.nspname AS schema,
    c.relname AS name,
    CASE WHEN c.relkind = 'r' THEN 'table' ELSE 'index' END AS type,
    pg_table_size(c.oid) AS size_bytes,
    (case WHEN reltuples < 0 THEN 0 else reltuples END)::int as row_estimate
FROM
    pg_class c
LEFT JOIN
    pg_namespace n ON n.oid = c.relnamespace
WHERE
    n.nspname NOT IN ('pg_catalog', 'information_schema')
    AND n.nspname !~ '^pg_toast'
    AND c.relkind IN ('r', 'i')
ORDER BY
    pg_table_size(c.oid) DESC,
    2 ASC
"""
        )
        return [RelationSpaceStat(*row) for row in cursor.fetchall()]


def get_table_sizes() -> list[RelationSpaceStat]:
    return [r for r in get_relation_sizes() if r.type == "table"]


def get_index_sizes() -> list[RelationSpaceStat]:
    return [r for r in get_relation_sizes() if r.type == "index"]



# Generated by Django 4.2.9 on 2024-01-17 03:19
import logging

from django.db import OperationalError, connection, migrations


def create_extension(apps, schema_editor):
    try:
        with connection.cursor() as cursor:
            cursor.execute("select extversion from pg_extension where extname = 'pg_stat_statements'")
            version = float(cursor.fetchone()[0])
            if version < 1.9:
                cursor.execute("alter extension pg_stat_statements update to '1.9'")
                cursor.execute("alter extension pg_stat_statements update to '1.10'")
    except OperationalError:
        logging.error("Could not update extension `pg_stat_statements`.")


class Migration(migrations.Migration):
    dependencies = [
        ("django_pev", "0001_enable_extensions"),
    ]

    operations = []



import datetime
import functools
import logging
import time
import traceback
import uuid
import webbrowser
from contextlib import contextmanager
from dataclasses import dataclass, field
from threading import local
from typing import Any

import sqlparse  # type:ignore[import]
from django.db import connections
from django.db.backends.utils import CursorWrapper
from django.utils import timezone

from django_pev.dalibo import PevResponse, upload_sql_plan
from django_pev.exceptions import PevException

logger = logging.Logger("django_pev")


thread_local_query_count = local()

CursorWrapper._original_execute = CursorWrapper._execute  # type:ignore
CursorWrapper._original_executemany = CursorWrapper._executemany  # type:ignore


class record_sql:
    """Record the SQL query and parameters for use in the explain view"""

    def __init__(self, cursor_wrapper: CursorWrapper, sql: str, params: Any):
        """Record the SQL query and parameters for use in the explain view"""
        self.cursor_wrapper = cursor_wrapper
        self.sql = sql
        self.params = params
        self.start_time = time.time()
        self.stack_trace = ""

    def __enter__(self):
        thread_local_query_count.query_count = getattr(thread_local_query_count, "query_count", 0) + 1
        frame_stack = traceback.extract_stack(limit=100)
        filtered_frame_stack = [
            x for x in frame_stack if not ("django_pev" in x.filename or "django/db" in x.filename)
        ][-10:]
        self.stack_trace = "".join(traceback.format_list(filtered_frame_stack))

    def __exit__(self, exc_type, exc_val, exc_tb):
        duration = time.time() - self.start_time
        thread_local_query_count.queries = getattr(thread_local_query_count, "queries", []) + [
            {
                "time": duration,
                "sql": self.cursor_wrapper.cursor.mogrify(self.sql, self.params).decode("utf-8"),
                "stack_trace": self.stack_trace,
            }
        ]


def _new_execute(self, sql, params, *ignored_wrapper_args):  # type: ignore[no-untyped-def] # fmt: skip
    with record_sql(self, sql, params):
        return CursorWrapper._original_execute(self, sql, params, *ignored_wrapper_args)


def _new_executemany(self, sql, params, *ignored_wrapper_args):  # type: ignore[no-untyped-def] # fmt: skip
    with record_sql(sql, params):
        return CursorWrapper._original_executemany(self, sql, params, *ignored_wrapper_args)


@dataclass(frozen=True)
class Explain:
    index: int
    duration: float
    sql: str
    stack_trace: str
    db_alias: str

    def __repr__(self) -> str:
        return f"Explain(duration={self.duration} sql={self.sql[:20]})"

    def __str__(self) -> str:
        return f"Explain(duration={self.duration} sql={self.sql[:20]})"

    @functools.cache  # noqa
    def visualize(self, upload_query: bool = False, analyze: bool = True, title: str = "") -> PevResponse:
        """Uploads the query and plan to explain.dalibo

        By default we do not embed the SQL query unless `upload_query` is set to True.

        The PevResponse object can be used to delete the uploaded plan.
        """
        with connections[self.db_alias].cursor() as cursor:
            if analyze:
                sql = f"EXPLAIN (ANALYZE, COSTS, VERBOSE, BUFFERS, FORMAT JSON) {self.sql}"
            else:
                sql = f"EXPLAIN (VERBOSE, FORMAT JSON) {self.sql}"
            cursor.execute(sql)
            plan = cursor.fetchone()[0]
        response = upload_sql_plan(query=self.sql if upload_query else "", plan=plan, title=title)
        logging.info(f"View Postgresql Explain @ {response.url}")
        return response

    def visualize_in_browser(self, upload_query: bool = False, analyze: bool = True, title: str = "") -> PevResponse:
        """Uploads the query and plan t oexplain.dalibo and then open in the browser"""
        response = self.visualize(upload_query, analyze, title)
        webbrowser.open(response.url)
        return response

    def explain(self, analyze: bool = True) -> str:
        """Runs explain and returns the plan as a string"""
        with connections[self.db_alias].cursor() as cursor:
            if analyze:
                sql = f"EXPLAIN (ANALYZE, COSTS, VERBOSE, BUFFERS) {self.sql}"
            else:
                sql = f"EXPLAIN (VERBOSE) {self.sql}"
            cursor.execute(sql)
            plan = "\n".join(list(x[0] for x in cursor.fetchall()))
        return plan


@dataclass
class ExplainSet:
    id: str
    url: str
    created: datetime.datetime
    queries: list[Explain] = field(default_factory=list)

    @property
    def n_queries(self) -> int:
        return len(self.queries)

    def __getitem__(self, index: int) -> Explain:
        return self.queries[index]

    @property
    def slowest(self) -> Explain:
        if not self.queries:
            raise PevException("Can not visualize results when there are no results.")

        return sorted(self.queries, key=lambda q: q.duration, reverse=True)[0]


@contextmanager
def explain(
    db_alias: str = "default",
    trace_limit: int = 10,
    url: str = "",
):
    """Capture all queries within this context and returns an ExplainSet container.

    Captured queries contain timing information, a stack trace and allows for visualization
    with explain.dalibo.com

    Usage:
    >>> with explain() as queries:
    >>>    User.objects.count()
    >>> result = queries.slowest.visualize(upload_query=True)
    >>> result.delete()
    >>> queries.slowest.explain()
    """
    result = ExplainSet(url=url, created=timezone.now(), id=str(uuid.uuid4()))
    thread_local_query_count.queries = []
    try:
        CursorWrapper._execute = _new_execute  # type:ignore # noqa
        CursorWrapper._executemany = _new_executemany  # type:ignore
        yield result
    finally:
        CursorWrapper._execute = CursorWrapper._original_execute  # type:ignore
        CursorWrapper._executemany = CursorWrapper._original_executemany  # type:ignore
        queries_after = thread_local_query_count.queries.copy()
    logger.debug(f"Captured {len(queries_after)} queries")

    for index, q in enumerate(queries_after):
        result.queries.append(
            Explain(
                index=index,
                duration=float(q["time"]),
                sql=sqlparse.format(q["sql"], reindent=True, keyword_case="upper"),
                stack_trace=q["stack_trace"],
                db_alias=db_alias,
            )
        )



from django.urls import path

from . import views

urlpatterns = [
    path("", views.QueriesView.as_view(), name="index"),
    path("space", views.SpaceView.as_view(), name="space"),
    path("maintenance", views.MaintenanceView.as_view(), name="maintenance"),
    path("connections", views.ConnectionsView.as_view(), name="connections"),
    path("indexes", views.IndexesView.as_view(), name="indexes"),
    path("live-queries", views.LiveQueriesView.as_view(), name="live-queries"),
    path("queries", views.QueriesView.as_view(), name="queries"),
    path("explain", views.ExplainView.as_view(), name="explain"),
    path("explain-visualize", views.ExplainVisualize.as_view(), name="explain-visualize"),
    path("embedded-pev", views.EmbeddedPev.as_view(), name="embedded-pev"),
]



import dataclasses
from datetime import datetime

from django.db import OperationalError, connection

TIME_DURATION_UNITS = (
    ("week", 60 * 60 * 24 * 7 * 1000),
    ("day", 60 * 60 * 24 * 1000),
    ("hour", 60 * 60 * 1000),
    ("min", 60 * 1000),
    ("sec", 1000),
    ("ms", 1),
)


def human_time_duration(milliseconds: float) -> str:
    if milliseconds == 0:
        return "inf"
    parts = []
    for unit, div in TIME_DURATION_UNITS:
        amount, milliseconds = divmod(int(milliseconds), div)
        if amount > 0:
            parts.append("{} {}".format(amount, unit))
    return " ".join(parts[:2])


@dataclasses.dataclass
class LiveQueryInfo:
    pid: int
    database: str
    state: str
    source: str
    duration: int
    waiting: bool
    query: str
    started_at: datetime
    duration_ms: float
    user: str
    backend_type: str

    @property
    def is_long_running(self) -> bool:
        if self.duration_ms:
            return self.duration_ms > 1000
        return False


def get_live_queries_current_database() -> list[LiveQueryInfo]:
    """Returns list of live queries"""
    sql = """
SELECT
    pid,
    datname as database,
    state,
    application_name AS source,
    age(NOW(), COALESCE(query_start, xact_start)) AS duration,
    wait_event is not null as waiting,
    query,
    COALESCE(query_start, xact_start) AS started_at,
    EXTRACT(EPOCH FROM NOW() - COALESCE(query_start, xact_start)) * 1000.0 AS duration_ms,
    usename AS user,
    backend_type
FROM
    pg_stat_activity
WHERE
    pid <> pg_backend_pid()
and
    datname = current_database()
ORDER BY
COALESCE(query_start, xact_start) DESC
"""

    with connection.cursor() as cursor:
        cursor.execute(sql)
        return list(LiveQueryInfo(*c) for c in cursor.fetchall())


@dataclasses.dataclass
class QueryStatInfo:
    query: str
    query_id: str
    query_md5: str
    user: str
    total_time: float
    mean_time: float
    stddev_time: float
    shared_blks_hit: int
    shared_blks_dirtied: int
    blk_read_time: int
    temp_blks_written: int
    temp_blks_read: int
    rows: int
    calls: int
    total_percent_time: float
    total_percent_calls: float
    total_percent_shared_blks: float
    calls_per_second: float
    shared_blks_per_second: float
    rows_per_second: float
    rows_per_call: float
    shared_blks_per_call: float

    @property
    def total_time_formatted(self) -> str:
        return human_time_duration(self.total_time)

    @property
    def mean_time_formatted(self) -> str:
        return human_time_duration(self.mean_time)

    @property
    def stddev_time_formatted(self) -> str:
        return human_time_duration(self.stddev_time)


def is_pg_stat_statements_installed() -> bool:
    try:
        with connection.cursor() as cursor:
            cursor.execute("SELECT 1 FROM pg_extension WHERE extname = 'pg_stat_statements'")
            assert cursor.fetchone()
            cursor.execute("SELECT 1 FROM pg_stat_statements limit 1")
            assert cursor.fetchone()
        return True
    except (AssertionError, OperationalError):
        return False


def enable_pg_stat_statements():
    with connection.cursor() as cursor:
        cursor.execute("CREATE EXTENSION IF NOT EXISTS pg_stat_statements")


def reset_pg_stat_statements():
    with connection.cursor() as cursor:
        cursor.execute("SELECT pg_stat_statements_reset()")


def get_query_stats() -> list[QueryStatInfo]:
    if not is_pg_stat_statements_installed():
        return []

    sql = """
WITH query_stats AS (
    SELECT
        LEFT(query, 10000) AS query,
        queryid as query_id,
        md5(query) as query_md5,
        rolname AS user,
        (total_exec_time) as total_time,
        mean_exec_time as mean_time,
        stddev_exec_time as stddev_time,
        shared_blks_hit,
        shared_blks_dirtied,
        blk_read_time,
        temp_blks_written,
        temp_blks_read,
        rows,
        calls
    FROM
        pg_stat_statements
    INNER JOIN
        pg_database ON pg_database.oid = pg_stat_statements.dbid
    INNER JOIN
        pg_roles ON pg_roles.oid = pg_stat_statements.userid
    WHERE
        calls > 0
    AND
        pg_database.datname = current_database()
), totals AS (
    SELECT
        SUM(total_exec_time) as total_time,
        SUM(calls) as total_calls,
        SUM(shared_blks_hit) as total_shared_blks_hit,
        SUM(shared_blks_dirtied) as total_shared_blks_dirtied,
        (extract(epoch from now()) - (select extract( epoch from stats_reset) from pg_stat_statements_info)) as total_seconds
    FROM pg_stat_statements
)
SELECT
    query,
    query_id,
    query_md5,
    query_stats.user,
    query_stats.total_time,
    mean_time,
    stddev_time,
    shared_blks_hit,
    shared_blks_dirtied,
    blk_read_time,
    temp_blks_written,
    temp_blks_read,
    rows,
    calls,
    query_stats.total_time * 100.0  / totals.total_time AS total_percent_time,
    calls * 100.0 / total_calls as total_percent_calls,
    (shared_blks_hit + shared_blks_dirtied) * 100.0 / (totals.total_shared_blks_hit + totals.total_shared_blks_dirtied) as total_percent_shared_blks,
    calls::float / total_seconds as calls_per_second,
    (shared_blks_hit + shared_blks_dirtied)::float / total_seconds as shared_blks_per_second,
    rows::float / total_seconds as rows_per_second,
    (shared_blks_hit + shared_blks_dirtied)::float / calls as shared_blks_per_call,
    rows::float/ calls as rows_per_call
FROM
    query_stats
CROSS JOIN totals
ORDER BY query_stats.TOTAL_TIME desc
LIMIT 100
"""
    with connection.cursor() as cursor:
        cursor.execute(sql)
        return list(QueryStatInfo(*c) for c in cursor.fetchall())
